/*** ThermoPC Linux data collection application. ***/
/*** (C)2002 BiPOM Electronics, Inc. All Rights Reserved ***/

/* Permission is granted to modify and use this program in your personal 
applications and in commercial applications only with products 
from BiPOM Electronics, Inc. */

/* Version history

Version 1.03

30 Mar 2004 - OMM Added command line arguments for verbose, log freq, units, help
			- OMM Added time display
			- OMM Added logging feature. For now it logs to a ixed file ( thermopc.log )

*/


#define VERSION		"1.03"


#include <stdio.h>
#include <time.h>
#include <asm/io.h>
#include <errno.h>

#define SUCCESS	0
#define ERRCODE unsigned int
#define FALSE 0
#define TRUE 1


#define UBYTE unsigned char
#define ULONG unsigned long
#define USHORT unsigned short
#define SHORT short
#define BOOL int


#define CELCIUS		0
#define FAHRENHEIT	1
#define KELVIN		2

#include "ds1621.h"



static void Nak( void );
static void Ack( void );
static ERRCODE Start( void );
static void Stop( void );

static UBYTE Receive( void );
static ERRCODE Transmit( UBYTE data );

static ERRCODE WritePort( UBYTE deviceNumber, UBYTE data );
static ERRCODE ReadTemperature( UBYTE deviceNumber, USHORT* rcvByte );
static ERRCODE ReadCounter( UBYTE deviceNumber, UBYTE* pCounter );
static ERRCODE ReadSlope( UBYTE deviceNumber, UBYTE* pSlope );
static ERRCODE ReadCfg( UBYTE deviceNumber, UBYTE* pCfg );
static ERRCODE WriteCfg( UBYTE deviceNumber, UBYTE cfg );
static ERRCODE StartConversion( UBYTE deviceNumber );
static ERRCODE ConvertCelsius( USHORT data, SHORT* pTemperature );
static void DelayMsec(ULONG milliseconds);
static void DelaySec(ULONG seconds);

static void setClk( void );
static void resetClk( void );
static void setData( void );
static void resetData( void );
static int getData( void );

static void applyPower();
static void setAddress();


void Output( USHORT port, UBYTE data );
UBYTE Input( USHORT port );

void Usage(void);
char* CurrTimeStr(void);


static USHORT port;

static char* UnitStr[] = { "CELCIUS", "FAHRENHEIT", "KELVIN" };




main(int argc, char* argv[])
{
	ERRCODE ec = SUCCESS;
	int i;
	USHORT temp;
	long logFrequency;
	int retVal;
	float convertedTemperature;
	BOOL verbose;
	int units;
	
	time_t currTime;
	time_t lastTime;
	
	FILE* fpLog = NULL;
	
	char* pFreqStr;
	char buffer[2048];
	
	//Fixed LPT1 port address
	port = 0x378;
	logFrequency = -1L;
	units = CELCIUS;
	verbose = FALSE;
	
	// Obtain permission to access port, port+1 and port+2 ( 3 bytes )
	retVal = ioperm( port, 3, 1 );
	
	//printf( "\nretval from ioperm=%d", retVal );

	if( retVal < 0 )
	{
		printf( "\nerrno=%d", errno );
		printf( "\nError setting I/O Port access permissions" );
		printf( "\nMake sure that you have root permissions" );
		goto func_end;
	}

	if( argc > 5 )
	{
		Usage();
		goto func_end;
	}

	for( i=0; i<argc; i++ )
	{
		if( !strcmp( argv[i], "/?" ) || !strcmp( argv[i], "--help" ) )
		{
			Usage();
			goto func_end;
		}
	}
	
	for( i=0; i<argc; i++ )
	{
		if( !strncmp( argv[i], "-l", 2 ) )
		{
			pFreqStr = ((char*)(argv[i]))+2;
			
			if( strlen(pFreqStr) )
			{
				logFrequency = atol(pFreqStr);				
			}
			else 
			{
				// Disable logging if a valid log requency is not supplied.
				logFrequency = -1L;
			}	
		}
		
		if( !strcmp( argv[i], "-f" ))
		{
			units = FAHRENHEIT;
		}
		else if( !strcmp( argv[i], "-k" ))
		{
			units = KELVIN;
		}

		if( !strcmp( argv[i], "-v" ))
		{
			verbose = TRUE;
		}
	}

	if( verbose )
	{
		printf( "\n*** Starting Temperature measurements ***" );
		
		if( logFrequency > 0 )
		{
			printf( "\nUnits=%s, logging=ON, log frequency=%d",
				UnitStr[units],
				logFrequency );
		}
		else
		{
			printf( "\nUnits=%s, logging=OFF",
				UnitStr[units],
				logFrequency );
		}
	}
	
	currTime = time(NULL);
	
	/* Read temperature values forever */
	for(;;)
	{
		ec = Temperature( port, (short*)&temp );
	
		// Print the temperature value ( in Celcius ) to console
		switch(units)
		{
		case FAHRENHEIT:
			convertedTemperature = ((float)temp)/10.0;
			convertedTemperature = (convertedTemperature * 9/5) + 32;
			sprintf( buffer, "\n%4.1f F %s", convertedTemperature, CurrTimeStr() );
			break;
			
		case KELVIN:
			convertedTemperature = ((float)temp)/10.0;
			convertedTemperature = convertedTemperature + 273.15; 
			sprintf( buffer, "\n%4.1f K %s", convertedTemperature, CurrTimeStr() );
			break;
			
		case CELCIUS:
		default:
			convertedTemperature = ((float)temp)/10.0;
			sprintf( buffer, "\n%4.1f C %s", convertedTemperature, CurrTimeStr() );
			break;
		}
		
		printf( "%s", buffer);
			
		//Logging: To be implemented
		if( logFrequency > 0 )
		{
			currTime = time(NULL);
			
			// Log if the time has come
			if( currTime >= logFrequency+lastTime )
			{
				lastTime = currTime;
				//printf( "\ntime=%d", time(NULL) );
				
				fpLog = fopen( "thermopc.log", "a+" );
				
				if( fpLog == NULL )
				{
					printf( "\nError opening log file" );
				}
				
				fprintf( fpLog, "%s", buffer );
				
				if( fpLog )
					fclose(fpLog);
			}
			
			//Base this on log frequency
			//t = CTime::GetCurrentTime();

			//timeStr = t.Format( "%c" );
			

			//sprintf( buffer, "%s,%d,C\n", timeStr,temp );

		}
	}

func_end:
	return 0;
}

ERRCODE Temperature( short lptPort, short* pTemp )
{
	ERRCODE ec = SUCCESS;

	UBYTE cfgReg;
	USHORT data;
	USHORT temp;
	UBYTE countRemain;
	UBYTE countPerC;

	float temperature;

	port = lptPort;

	applyPower();

	setAddress();

	// Get the configuration from temperature sensor
	ec = ReadCfg( DS1621_ADDR, &cfgReg );

    if( ec != SUCCESS )
        goto func_end;

	// If not already in one-shot mode, switch to one-shot mode 
	if( !(cfgReg & 0x01) )
	{
		cfgReg |= 0x01;

		ec = WriteCfg( DS1621_ADDR, cfgReg );

		DelayMsec(100);

        if( ec != SUCCESS )
            goto func_end;
	}

	// Start temperature conversion
	ec = StartConversion( DS1621_ADDR );

    if( ec != SUCCESS )
        goto func_end;

	// Delay 1 sec until readings are stabilized
	DelaySec(1);

	ec = ReadCfg( DS1621_ADDR, &cfgReg );

    if( ec != SUCCESS )
        goto func_end;

	// Read raw temperature value
	ec = ReadTemperature( DS1621_ADDR, &data );

    if( ec != SUCCESS )
        goto func_end;

	// Read counter remainder
	ec = ReadCounter( DS1621_ADDR, &countRemain );
	
    if( ec != SUCCESS )
        goto func_end;

	// Read the slope 
	ec = ReadSlope( DS1621_ADDR, &countPerC );
	
    if( ec != SUCCESS )
        goto func_end;

	temp = data >> 8;

	if( countPerC != 0 )
		temperature = (float)temp - 0.25 + ( countPerC - countRemain ) / (float)countPerC;
	else
		temperature = (float)temp - 0.25;

func_end:

    if( ec == SUCCESS )
    {
        double fraction;
        double integer;

	temperature *= 10;

        fraction = modf( temperature, &integer );

        // round off the number
        if( fraction >= 0.5 )
        {
            integer++;
        }
        else if( fraction <= -0.5 )
        {
            integer--;
        }


	    *pTemp = (short)integer;
    }
    else
    {
        *pTemp = 0;
    }

	return ec;
}





static ERRCODE ConvertCelsius( USHORT data, SHORT* pTemperature )
{
	ERRCODE ec = SUCCESS;

	*pTemperature = data >> 8;

	return ec;
}




/*********************************************************************
*                                                                   
*      WritePort - Writes a data byte to the specified address
*                 in EEPROM.
*
*      Inputs:  UBYTE  - device number on the bus
*               unsigned       - 16-bit address
*               UBYTE - data byte to write
*
*      Outputs -  Address to EEPROM
*
*      Returns -  ERRCODE - SUCCESS if operation successful
*                           non-zero error code otherwise
*                                                                            
*********************************************************************/

static ERRCODE WritePort( UBYTE deviceNumber, UBYTE data )
{
    ERRCODE ec = SUCCESS;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( data );

    if( ec != SUCCESS )
        goto func_end;

func_end:

    Stop();

    return ec;
}


static ERRCODE WriteCfg( UBYTE deviceNumber, UBYTE cfg )
{
    ERRCODE ec = SUCCESS;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

   ec = Transmit( CMD_WRITE_CFG );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( cfg );

    if( ec != SUCCESS )
        goto func_end;

func_end:

    Stop();

    return ec;
}



/*********************************************************************
*                                                                   
*      ReadPort - Reads a data byte from the port on the specified 
*		  PCF8574 I/O Expander.
*
*      Inputs:  unsigned char  - device number on the bus
*               unsigned char* - where to put error code
*
*      Outputs -  Address to EEPROM
*
*      Returns -  data byte
*                                                                                       
*********************************************************************/

static ERRCODE ReadTemperature( UBYTE deviceNumber, USHORT* pTemperature )
{
    ERRCODE ec = SUCCESS;
    UBYTE temp = 0;

	UBYTE msb = 0;
	UBYTE lsb = 0;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( CMD_READ_LAST_TEMPERATURE );

    if( ec != SUCCESS )
        goto func_end;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_READ | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    msb = Receive();

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_READ | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    lsb = Receive();

    Nak();

func_end:

    Stop();

    *pTemperature = ( msb << 8 ) | lsb;

    return ec;
}


static ERRCODE StartConversion( UBYTE deviceNumber )
{
    ERRCODE ec = SUCCESS;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( CMD_START_CONVERSION );

    if( ec != SUCCESS )
        goto func_end;


func_end:

    Stop();

    return ec;
}



static ERRCODE ReadCounter( UBYTE deviceNumber, UBYTE* pCounter )
{
    ERRCODE ec = SUCCESS;
    UBYTE temp = 0;

	UBYTE counter = 0;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( CMD_READ_COUNTER );

    if( ec != SUCCESS )
        goto func_end;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_READ | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    counter = Receive();

    Nak();

func_end:

    Stop();

    *pCounter = counter;

    return ec;
}



static ERRCODE ReadSlope( UBYTE deviceNumber, UBYTE* pSlope )
{
    ERRCODE ec = SUCCESS;
    UBYTE temp = 0;

	UBYTE slope = 0;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( CMD_READ_SLOPE );

    if( ec != SUCCESS )
        goto func_end;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_READ | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    slope = Receive();

    Nak();

func_end:

    Stop();

    *pSlope = slope;

    return ec;
}



static ERRCODE ReadCfg( UBYTE deviceNumber, UBYTE* pCfg )
{
    ERRCODE ec = SUCCESS;

	UBYTE cfg = 0;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_WRITE | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    ec = Transmit( CMD_READ_CFG );

    if( ec != SUCCESS )
        goto func_end;

    ec = Start();

    if( ec != SUCCESS )
        return ec;

    ec = Transmit( I2C_DEVICE_DS1621 | I2C_READ | (deviceNumber << 1) );

    if( ec != SUCCESS )
        goto func_end;

    cfg = Receive();

    Nak();

func_end:

    Stop();

    *pCfg = cfg;

    return ec;
}



/*
; Function: Transmit
;
; Description: Transmits a byte to the EEPROM with 
;			   most significant bit first.
;
; Inputs: 	 A - Contains the byte to be transmitted.
;			 SCL and SDA should be low on entry.
;
; Returns:	 - Carry is set if EEPROM can not acknowledge.
;			 - A is modified.
*/
static ERRCODE Transmit( UBYTE data )
{
	ERRCODE ec = SUCCESS;

	int i;

	for( i=0; i<8; i++ )
	{
		if( data & 0x80 )
		{
			setData();
		}
		else
		{
			resetData();
		}

		setClk();

		resetClk();

		data = data << 1;
	}
	
	setData();
	setData();

	setClk();

	resetClk();

func_end:

	return ec;
}



/*
; Function: Receive
;
; Description: Receives a byte from the EEPROM with 
;			   most significant bit first.
;
; Inputs: 	 A - Contains the byte to be transmitted.
;			 SCL should be low on entry.
;
; Returns:	 SCL is reset upon return.
;			 A contains the received byte.
*/
static UBYTE Receive( void )
{
	int i;

	UBYTE data = 0;

	for( i=0; i<8; i++ )
	{
		data = data << 1;

		setClk();

		if( getData() )
		{
			data |= 0x01;
		}

		resetClk();

	}

	return data;
}


/*
; Function: Nak
;
; Description: Signals a negative acknowledgement 
;			   by clocking data (SDA) high
;
; Inputs: 	 SCL should be low upon entry
;
; Returns:	 SCL is low; SDA is high upon return.
*/
static void Nak( void )
{
	/* Negative acknowledgement */
	setData();
	
	setClk();

	resetClk();
}


/*
; Function: Ack
;
; Description: Signals an acknowledgement 
;			   by clocking data (SDA) low
;
; Inputs: 	 SCL should be low upon entry
;
; Returns:	 SCL and SDA are high upon return.
*/
static void Ack( void )
{
	/* Acknowledge by pulling data low */
	resetData();

	setClk();

    resetClk();
}


/*
; Function: Start
;
; Description: Signals a Start which is defined as negative going transition
;			   of SDA while SCL is high.
;
; Inputs: 	 None
;
; Returns:	 - Carry is set if bus is not available.
;			 - Carry is reset if success.
;			 - SCL and SDA are reset upon return.
*/
static ERRCODE Start( void )
{
	ERRCODE ec = SUCCESS;

	/* Set both SCL and SDA high */
	setData();

	setClk();

	/* omm Check if the bus is available; check only the data pin */

    // return error here if bus is not available
	
	/* Reset SDA first, then SCL to signal a Start */

	resetData();
	resetData();
	
	resetClk();    

func_end:

	return ec;
}



/*
; Function: Stop
;
; Description: Signals a Stop which is defined as positive going transition
;			   of SDA while SCL is high.
;
; Inputs: 	 SCL should be reset on entry
;
; Returns:	 SCL and SDA are set upon return.
*/
static void Stop( void )
{
	resetData();

	setClk();

	setData();
}


static void setClk( void )
{
    Output( port, Input(port) | 0x80 );
	DelayMsec(1);
}

static void resetClk( void )
{
    Output( port, Input(port) & 0x7F );
	DelayMsec(1);
}


static void setData( void )
{
	/* Inverse logic; reset register bit to set port bit */
	Output( port+2, Input(port+2) & 0xF7 );
	DelayMsec(1);
}

static void resetData( void )
{
	/* Inverse logic; set register bit to reset port bit */
	Output( port+2, Input(port+2) | 0x08 );
	DelayMsec(1);
}

static int getData( void )
{
	int temp;

	/* Data line should be HIGH so we can read it */
	setData();


	temp = Input( port+1 );

	/* Return 0 if port pin is reset, non-zero if port pin is set */
	/* Note that register bit is reverse of port pin */
	if( temp & 0x10 )
		return 0;
	else
		return 1;

}



static void applyPower()
{
	// Turn ON D7 and D6, STROBE should already be ON by default
	Output( port, Input(port) | 0x80 | 0x40 | 0x20 | 0x10 | 0x07 );
	setData();

}


static void setAddress()
{
	//TODO set status of STROBE as well

	// Turn ON D7 and D6, STROBE should already be ON by default
	Output( port, Input(port) | 0x80 | 0x40 );

	// Turn OFF D5
	Output( port, Input(port) & 0xDF );

	DelayMsec(1);
}




/* OS Specific Code */
void DelaySec(ULONG seconds)
{
	sleep(seconds);
}

void DelayMsec(ULONG milliseconds)
{
	struct timespec req;
	ULONG i;
	ULONG microseconds;

	microseconds = milliseconds*1000;

	//Use a port read to generate a approximate 1 usec delay 
	for( i=0; i<microseconds; i++ )
		inb(port);
}


void Output( USHORT port, UBYTE data )
{
	//printf( "\nwriting %02x to port %04x", data, port );	
	outb(data,port);
}

UBYTE Input( USHORT port )
{
	return inb(port);
}


void Usage()
{
	printf( "\nUsage:");
	printf( "\n" );
	printf( "\nthermopc_unix -l<log frequency> -k -f -v --help" );
	printf( "\nwhere log frequency is in seconds" );
	printf( "\n-k=unit is KELVIN" );
	printf( "\n-f=unit is FAHRENHEIT" );
	printf( "\n-v=enable VERBOSE" );
	printf( "\n--help = display Usage" );
	printf( "\nDefault unit is CELCIUS" );
}


char* CurrTimeStr(void)
{
	static char timeBuffer[256];
	
	time_t timer;

  	timer=time(NULL);
  	
  	strftime(timeBuffer, sizeof(timeBuffer), "%H:%M:%S %m/%d/%Y", localtime(&timer));
  	
  	return timeBuffer;
}

/*EOF*/
